from overrides import overrides
from typing import Dict, Tuple, List

import torch

from allennlp.modules import Attention
from allennlp.modules.bimpm_matching  import BiMpmMatching

from src.tensor_op import sorted_dynamic_parition, dense2sparse, sparse2dense
from .graph_pair2graph_pair_encoder import GraphPair2GraphPairEncoder


@GraphPair2GraphPairEncoder.register("bimpm", exist_ok=True)
class GraphPairMPM(GraphPair2GraphPairEncoder):
    """
    BiMPM for Sparse Batch By Dense, Sparse Convertion
    """
    def __init__(self,
                 bimpm: BiMpmMatching,
                ):
        super(GraphPairMPM, self).__init__()
        self._bimpm = bimpm
        self._dim_match = bimpm.get_output_dim()
        
    def get_output_dim():
        return self._dim_match
    
    def forward(self,
                x1: torch.Tensor,x2: torch.Tensor,
                b1: torch.Tensor,
                b2: torch.Tensor,
                return_attention: bool = False,
    ) -> Tuple[torch.Tensor]:
        x1 = sparse2dense(x1, b1)
        x2 = sparse2dense(x2, b2)
        x1, m1 = x1["data"], x1["mask"]
        x2, m2 = x2["data"], x2["mask"]
        x1, x2 = self._bimpm(x1, m1, x2, m2) 
        x1 = torch.cat(x1, dim=-1)
        x2 = torch.cat(x2, dim=-1)
        # concat different view (max/mean/attentive/attentivemax)
        # can try use attentive max + attentive only
        x1 = dense2sparse(x1, m1)
        x2 = dense2sparse(x2, m2)
        if return_attention:
            return x1["data"], x2["data"], None
        else:
            return x1["data"], x2["data"]
    
    def __repr__(self):
        return f"({str(self._bimpm)}, dim_matching_output={str(self._dim_match)})"
    
    
"""
BiMPM
    hidden_dim: int = 100,
    num_perspectives: int = 20,
    share_weights_between_directions: bool = True,
    is_forward: bool = None,
    with_full_match: bool = True,
    with_maxpool_match: bool = True,
    with_attentive_match: bool = True,
    with_max_attentive_match: bool = True,
"""

"""
def forward(self,
           x1: torch.Tensor,
           x2: torch.Tensor,
           b1: torch.Tensor,
           b2: torch.Tensor) -> Tuple[torch.Tensor]:
    # N is number of batches
    N: int = b1[-1]+1
    x1: List[torch.Tensor] = sorted_dynamic_parition(x1, b1)
    x2: List[torch.Tensor] = sorted_dynamic_parition(x2, b2)
    # return all nodes
    r1 = []
    r2 = []
    # for each batch
    for i in range(N):
        bx1 = xs1[i] # n*d
        bx2 = xs2[i] # m*d
        S = self._att(bx1, bx2) # n*m
        r1 += [bx1]
        r2 += []

    r1 = torch.cat(r1, dim=0)
    r2 = torch.cat(r2, dim=0)

    return r1, r2
"""





from allennlp.common import Registrable # if add more 

from overrides import overrides
from typing import Dict, Tuple, List
from src.tensor_op import sorted_dynamic_parition

import torch


class GraphPair2GraphPairEncoder(torch.nn.Module, Registrable):
    pass


# to be solved "[ROOT]" is not special for transformer
# can use "[SEP]" ?, since we are not using [SEP], and [SEP] is not meaningful itself without "type_id"?
 
import src.data_git.reader_config as config
import src.data_git.utils as utils
from src.data_git.sparse_adjacency_field import SparseAdjacencyField

#import os, sys; sys.path.append(os.path.dirname(os.path.realpath(__file__)))
#from src_gmn.sparse_adjacency_field import SparseAdjacencyField

import itertools
from typing import Iterable, List, Dict, Tuple, Union
import json
import logging

from overrides import overrides

from allennlp.common.file_utils import cached_path
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.fields import Field, TextField, LabelField, MetadataField, ArrayField, AdjacencyField
from allennlp.data.instance import Instance
from allennlp.data.token_indexers import SingleIdTokenIndexer, TokenIndexer, PretrainedTransformerMismatchedIndexer
from allennlp.data.tokenizers import Tokenizer, SpacyTokenizer, PretrainedTransformerTokenizer
from allennlp.data import Token, Vocabulary

# pytorch geometric
from torch_geometric.utils.convert import to_networkx
from torch_geometric.data.data import Data as PyGeoData
from torch_geometric.data import DataLoader
## Stanza
import stanza
from stanza.models.common.doc import Document as StanzaDoc
from stanza.pipeline.core import Pipeline as StanzaPipeline

logger = logging.getLogger(__name__)

default_fields = ["sentence1", "sentence2", "gold_label"]

"""
Caveat:
    the file is current dependent on reader_config.py
    "[ROOT]" is not special token should be fixed
    combined_input_field is not implemented yet
"""

# comment if in development step for ipython notebook import
# or use @DatasetReader.register("nli-graph", exist_ok=True)
@DatasetReader.register("nli-graph", exist_ok=True)
class NLIGraphReader(DatasetReader):
    """
    Reads a file from a preprocessed/raw NLI dataset.
    the input type can be determined by __init__ parameter
    This data is formatted as jsonl, one json-formatted instance per line.
    {
        "gold_label": {0: contradiction, 1: neutral, 2: entailment} or RawLabel
        "sentence1": StanzaDoc in List form or RawText
        "sentence2": StanzaDoc in List form or RawText
    }
    along with a metadata field containing the tokenized strings of the premise and hypothesis.
    Registered as a `DatasetReader` with name "nli-graph".
    
    # Parameter:
        wordpiece_tokenizer: Tokenizer, optional (default=`{"tokens": PretrainedTransformerTokenizer(config.TRANSFORMER_NAME)}`)
            tokenize token into smaller pieces
        token_indexers : `Dict[str, TokenIndexer]`, optional (default=`{"tokens": PretrainedTransformerMismatchedIndexer(config.TRANSFORMER_NAME)}`)
            index token
        combine_input_fields : `bool`, optional(default=False)
            whether to encode P and H at the same time with [CLS]P[SEP]H[SEP]
            Note: if want to use BERT like NLI method, see original reader on github "allennlp-models/esim..."
        input_parsed: `bool`, optional (default=`True`)
            if the input is rawtext or parsed stanza doc
            if not, the parsing part is in text2instance function
            and the cache option for readsetreader should be open(cache_dir)
        parser: `StanzaPipeline`, optional
            if input_parsed is provided False, provide parser
        cache_directory : `str`, optional (default=`None`)
            this is a param for parent class `DatasetReader`
            from  `https://github.com/allenai/allennlp/blob/master/allennlp/data/dataset_readers/dataset_reader.py`
            If given, we will use this directory to store a cache of already-processed `Instances` in
            every file passed to :func:`read`, serialized (by default, though you can override this) as
            one string-formatted `Instance` per line.  If the cache file for a given `file_path` exists,
            we read the `Instances` from the cache instead of re-processing the data (using
            :func:`_instances_from_cache_file`).  If the cache file does _not_ exist, we will _create_
            it on our first pass through the data (using :func:`_instances_to_cache_file`).
            !!! NOTE
                It is the _caller's_ responsibility to make sure that this directory is
                unique for any combination of code and parameters that you use.  That is, if you pass a
                directory here, we will use any existing cache files in that directory _regardless of the
                parameters you set for this DatasetReader!_
    
    # Caveat:
        the file is current dependent on reader_config.py
        "[ROOT]" is not special token should be fixed
        combined_input_field is not implemented yet
    """

    def __init__(
        self,
        wordpiece_tokenizer: Tokenizer = None,
        token_indexers: Dict[str, TokenIndexer] = None,
        combine_input_fields: bool = None,
        input_parsed: bool = None,
        parser: StanzaPipeline = None,
        input_fields: List = None,
        **kwargs,
    ) -> None:
        #super().__init__(manual_distributed_sharding=True, **kwargs)
        super(NLIGraphReader, self).__init__(**kwargs)
        self._wordpiece_tokenizer = wordpiece_tokenizer # used when want to combine input field if use bert
        self._token_indexers = token_indexers 
        self._combine_input_fields = combine_input_fields or False
        self._input_parsed = input_parsed if (input_parsed is not None) else True 
        self._parser = parser or None
        self.f = input_fields or default_fields #remove dependency to config

    @overrides
    def _read(self, file_path: str):
        """
        Reads a file, yield instances
        can take raw or parsed depends on __init__ param (input_parsed)
        the parsing part is in text2instance function
        besure to open cache in config file to store instance if using raw text to speed up coming experiments 
        """
        file_path = cached_path(file_path)
        with open(file_path, "r") as fo:
            example_iter = (json.loads(line) for line in fo.readlines())
            # we need gold label
            filtered_example_iter = (
                example for example in example_iter if example[self.f[2]] != "-"
            )
            for example in filtered_example_iter:
                # we want label to be in string here, use vocab to label
                label = example[self.f[2]]
                if(isinstance(label, int)):
                    label = config.id_to_label[label]
                premise : Union[StanzaDoc, List, str] = example[self.f[0]]
                hypothesis :  Union[StanzaDoc, List, str] = example[self.f[1]]
                yield self.text_to_instance(premise, hypothesis, label)
                
    
    #@classmethod
    def instance2sent(self, instance):
        """
        return string for instance
        todo: add graph visualization
        """
        return {
            "premise": " ".join([token.text for token in instance.fields["tokens_p"].tokens]),
            "hypothesis": " ".join([token.text for token in instance.fields["tokens_h"].tokens]),
        }
        
    
    @overrides
    def text_to_instance(
        self, 
        premise: Union[StanzaDoc, List, str],
        hypothesis: Union[StanzaDoc, List, str],
        gold_label: Union[int, str]= None,
    ) -> Instance:
        """
        input: premise/hypothesis as List of Graph Infromation
        output: allennlp Instance
        
        convert List to PytorchGeo Data by utils.doc2graph
        node_attr : word tokens 
        edge_attr : edge labels
        edge_index : sparse edge
        """

        fields: Dict[str, Field] = {}
        
        if isinstance(premise, str): # or use ` not self._input_paresed`
            premise : StanzaDoc = self._parser(premise)
            hypothesis : StanzaDoc = self._parser(hypothesis)
        g_p: PyGeoData = utils.doc2graph(premise)
        g_h: PyGeoData = utils.doc2graph(hypothesis)
        def stanza_word2allennlp_token(w):
            t = Token(
                text = w.text,
                lemma_ = w.lemma,
                pos_ = w.pos,
                dep_ =  w.deprel,
            )
            return t 
        tokens_p: List[Token] = [stanza_word2allennlp_token(w) for w in  g_p.node_attr]
        tokens_h: List[Token] = [stanza_word2allennlp_token(w) for w in  g_h.node_attr]
            
        if self._combine_input_fields:
            # see the doc, the output should be indexed
            raise NotImplementedError
            tokens = self._wordpiece_tokenizer.add_special_tokens(tokens_p, tokens_h)
            fields["tokens"] = TextField(tokens, self._token_indexers)
        else:
            #premise_tokens = self._wordpiece_tokenizer.add_special_tokens(g_p.node_attr)
            #hypothesis_tokens = self._wordpiece_tokenizer.add_special_tokens(g_h.node_attr)
            fields["tokens_p"] = TextField(tokens_p, self._token_indexers) # defualt = {tokens: tokens}?
            fields["tokens_h"] = TextField(tokens_h, self._token_indexers)
        fields["g_p"] = SparseAdjacencyField(graph=g_p,
                                             sequence_field=fields["tokens_p"],
                                             label_namespace = "edge_labels"
                                            )
        fields["g_h"] = SparseAdjacencyField(graph=g_h,
                                             sequence_field=fields["tokens_h"],
                                             label_namespace = "edge_labels"
                                            )
        
        # care do not use `if label` for label can be 0(in int)
        if gold_label is not None:
            fields["label"] = LabelField(gold_label, skip_indexing=False, label_namespace="labels")

        return Instance(fields)


## util
import os
import logging
from argparse import ArgumentParser
from tqdm import tqdm_notebook as tqdmnb
from tqdm import tqdm as tqdm
import pickle
import json 
import jsonlines as jsonl
from collections import defaultdict
from typing import Iterable, List, Dict, Tuple, Union
from pathlib import Path
## graph
import networkx as nx
import matplotlib.pyplot as plt
## nn
import numpy as np
import torch
from torch_geometric.utils.convert import to_networkx
from torch_geometric.data.data import Data as PytorchGeoData
from torch_geometric.data import DataLoader as PytorchGeoDataLoader
## Stanza
import stanza
from stanza.models.common.doc import Document as StanzaDocument
from stanza.pipeline.core import Pipeline as StanzaPipeline
## allennlp model
from allennlp_models.structured_prediction.predictors.srl import SemanticRoleLabelerPredictor
from allennlp_models.structured_prediction.predictors.biaffine_dependency_parser import BiaffineDependencyParserPredictor
from allennlp.predictors.predictor import Predictor #

## self

###########################################################################################

# root_token
root_token = "$"

def g2sent(g : PytorchGeoData):
    return " ".join(g.node_attr)



def draw_edges(edge_index: torch.Tensor, tokens: List[str])->None:
    fig = plt.figure() # figsize=None
    ax = fig.add_subplot(111)
    cax = ax.imshow(att, cmap='bone', origin='upper')
    fig.colorbar(cax)
    # Set up axes
    ax.set_xticks(np.arange(len(seq)))
    ax.set_xticklabels(seq)
    ax.set_yticklabels([])
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")


    # Show label at every tick
    #ax.xaxis.set_major_locator(ticker.MultipleLocator(1))
    #ax.yaxis.set_major_locator(ticker.MultipleLocator(1))
    ax.set_title("Attention")
    fig.tight_layout()
    plt.show()
    return 

def draw(data : PytorchGeoData, node_size=1000, font_size=12, save_img_file=None):
    """
    input: (torch_geometric.data.data.Data, path or string)
    effect: show and save graph data, with graphviz layout visualization
    """
    G = to_networkx(data)
    pos = nx.nx_pydot.graphviz_layout(G)
    if(data.edge_attr != None):
        edge_labels = {(u.item(),v.item()):lab for u,v,lab in zip(data.edge_index[0], data.edge_index[1], data.edge_attr)}
    else:
        edge_labels = None
    if(data.node_attr != None):
        node_labels = dict(zip(G.nodes, data.node_attr))
    else:
        node_labels = None
    nx.draw(G, pos=pos, nodecolor='r', edge_color='b', node_size=node_size, with_labels=False)
    nx.draw_networkx_labels(G, pos=pos, labels=node_labels, font_size=font_size)
    nx.draw_networkx_edge_labels(G, pos=pos, edge_labels=edge_labels, font_size=font_size)
    print(G.nodes)
    print(G.edges)
    if save_img_file is not  None:
        plt.savefig(save_img_file)
    plt.show()
    return

######################################################################################################
# Stanza   #
######################################################################################################

def text2graph(text : str, nlp : StanzaPipeline) -> PytorchGeoData:
    """
    text2doc by Stanza
    doc2graph by utils.doc2graph 
    """
    return doc2graph(nlp(text))
    
def doc2graph(doc : Union[StanzaDocument, List]) -> PytorchGeoData:
    """
    input:
        doc : Union[StanzaDocument, List]
        selected_features : List[str] (this is to be added)
    output:
        PytorchGeoData : G
        G = {
         x: id tensor
         edge_idx : edges size = (2, l-1)
         edge_attr: (u, v, edge_type in str)
         node_attr: list of StanzaWords
        }
    Note that in Stanza, token is a list of words
    """
    if isinstance(doc, list): #convert to Doc first if is in dict form ([[dict]])
        doc = StanzaDocument(doc)
    # add root token for each sentences
    n = doc.num_words
    e = [[],[]]
    edge_info = []
    node_info = []
    prev_token_sum = 0
    # add edge function:
    def add_edge(id1, id2, type_, bidirectional=True):
        """
        add edge to list
        bidirectional?
        """
        e[0].append(id1)
        e[1].append(id2)
        edge_info.append(type_)
        if bidirectional:
            e[0].append(id2)
            e[1].append(id1)
            edge_info.append("reverse:"+type_)
    # get original dependency
    for idx, sent in enumerate(doc.sentences):
        # node info by index(add root at the beginning of every sentence)
        node_info.extend(sent.words)
        # edge info by index of u in edge (u,v)
        for dep in sent.dependencies:
            if dep[1] != "root":
                id1 = prev_token_sum + int(dep[0].id)-1
                id2 = prev_token_sum + int(dep[2].id)-1
                add_edge(id1, id2, dep[1])
        prev_token_sum += len(sent.words)
        # sent.print_dependencies()
        # sent.print_words()
    # add constituent edges
    for i in range(n-1):
        add_edge(i, i+1, "const:next", bidirectional=False)
        add_edge(i+1, i, "const:prev", bidirectional=False)
    # done building edges and nodes
    # print(n, e, edge_info, node_info, sep='\n')
    x = torch.tensor(list(range(n)))
    e = torch.LongTensor(e) # use longtensor explicitly so that Data can init without type issue
    G = PytorchGeoData(x=x, edge_index=e, edge_attr=edge_info, node_attr=node_info)
    return G



####################################################################################
# if direct use allen nlp dependencies predictor without sentence segmentation     #
####################################################################################
def text2graph_allennlp(text: str, dep_predictor : BiaffineDependencyParserPredictor) -> PytorchGeoData:
    """
    text2doc by nlp
    doc2graph by utils.doc2graph 
    """
    return doc2graph_allennlp(dep_predictor.predict(sentence=text))

# Stanza Parse Seems Stabler than Allennlp's...
def doc2graph_allennlp(doc: Dict) -> PytorchGeoData:
    """
    input: allen dependecies (Dict)
    return G = {
     x: id tensor
     edge_idx : edges size = (2, l-1)
     edge_attr: (u, v, edge_type in str)
     node_attr: text
    }
    """
    # add root token for each sentences
    n = len(doc["words"])
    e = [list(range(1, n+1)),doc["predicted_heads"]]
    edge_attr = list(zip(e[0], e[1], doc["predicted_dependencies"]))
    node_attr = [root_token]
    node_attr.extend(doc["words"])
    x = torch.tensor(list(range(n)))
    e = torch.tensor(e)
    G =PytorchGeoData(x=x, edge_index=e, edge_attr=edge_attr, node_attr=node_attr)
    return G


"""
# root version doc2graph
def doc2graph(doc : Union[StanzaDocument, List]) -> PytorchGeoData:
    if isinstance(doc, list): #convert to Doc first if is in dict form ([[dict]])
        doc = StanzaDocument(doc)
    # add root token for each sentences
    n = doc.num_tokens+len(doc.sentences)
    e = [[],[]]
    edge_info = []
    node_info = []
    prev_token_sum = 0
    prev_root_id = 0
    cur_root_id = 0
    # add edge function:
    def add_edge(id1, id2, type_, bidirectional=True):
        e[0].append(id1)
        e[1].append(id2)
        edge_info.append(type_)
        if bidirectional:
            e[0].append(id2)
            e[1].append(id1)
            edge_info.append("reverse:"+type_)
    # get original dependency
    for idx, sent in enumerate(doc.sentences):
        # node info by index(add root at the beginning of every sentence)
        cur_root_id = len(node_info)
        node_info.append(root_token)
        for token in sent.tokens:
            node_info.append(token.to_dict()[0]['text'])
        # edge info by index of u in edge (u,v)
        for dep in sent.dependencies:
            id1 = prev_token_sum + int(dep[0].to_dict()["id"]) # here no need to -1 because of first root
            id2 = prev_token_sum + int(dep[2].to_dict()["id"])
            add_edge(id1, id2, dep[1])
        prev_token_sum += len(sent.tokens)+1
        # add links between sentence roots
        if(cur_root_id != 0):
            id1 = prev_root_id
            id2 = cur_root_id
            add_edge(id1, id2, "bridge")
        prev_root_id = cur_root_id
    # done building edges and nodes
    x = torch.tensor(list(range(n)))
    e = torch.tensor(e)
    G = PytorchGeoData(x=x, edge_index=e, edge_attr=edge_info, node_attr=node_info)
    return G
"""


import sys
import inspect

import torch
import torch_scatter

special_args = [
    'edge_index', 'edge_index_i', 'edge_index_j', 'size', 'size_i', 'size_j'
]
__size_error_msg__ = ('All tensors which should get mapped to the same source '
                      'or target nodes must be of same size in dimension 0.')

is_python2 = sys.version_info[0] < 3
getargspec = inspect.getargspec if is_python2 else inspect.getfullargspec



def scatter_(name, src, index, dim=0, dim_size=None):
    r"""Aggregates all values from the :attr:`src` tensor at the indices
    specified in the :attr:`index` tensor along the first dimension.
    If multiple indices reference the same location, their contributions
    are aggregated according to :attr:`name` (either :obj:`"add"`,
    :obj:`"mean"` or :obj:`"max"`).
    Args:
        name (string): The aggregation to use (:obj:`"add"`, :obj:`"mean"`,
            :obj:`"max"`).
        src (Tensor): The source tensor.
        index (LongTensor): The indices of elements to scatter.
        dim (int, optional): The axis along which to index. (default: :obj:`0`)
        dim_size (int, optional): Automatically create output tensor with size
            :attr:`dim_size` in the first dimension. If set to :attr:`None`, a
            minimal sized output tensor is returned. (default: :obj:`None`)
    :rtype: :class:`Tensor`
    """

    assert name in ['add', 'mean', 'max']

    if name == 'max':
        op = torch.finfo if torch.is_floating_point(src) else torch.iinfo
        fill_value = op(src.dtype).min
    else:
        fill_value = 0

    op = getattr(torch_scatter, 'scatter_{}'.format(name))
    #out = op(src, index, dim, None, dim_size, fill_value)
    out = op(src, index, dim, dim_size, fill_value)
    if isinstance(out, tuple):
        out = out[0]

    if name == 'max':
        out[out == fill_value] = 0

    return out

class MessagePassing(torch.nn.Module):
    r"""Base class for creating message passing layers
    .. math::
        \mathbf{x}_i^{\prime} = \gamma_{\mathbf{\Theta}} \left( \mathbf{x}_i,
        \square_{j \in \mathcal{N}(i)} \, \phi_{\mathbf{\Theta}}
        \left(\mathbf{x}_i, \mathbf{x}_j,\mathbf{e}_{i,j}\right) \right),
    where :math:`\square` denotes a differentiable, permutation invariant
    function, *e.g.*, sum, mean or max, and :math:`\gamma_{\mathbf{\Theta}}`
    and :math:`\phi_{\mathbf{\Theta}}` denote differentiable functions such as
    MLPs.
    See `here <https://pytorch-geometric.readthedocs.io/en/latest/notes/
    create_gnn.html>`__ for the accompanying tutorial.
    Args:
        aggr (string, optional): The aggregation scheme to use
            (:obj:`"add"`, :obj:`"mean"` or :obj:`"max"`).
            (default: :obj:`"add"`)
        flow (string, optional): The flow direction of message passing
            (:obj:`"source_to_target"` or :obj:`"target_to_source"`).
            (default: :obj:`"source_to_target"`)
    """
    def __init__(self, aggr='add', flow='source_to_target'):
        super(MessagePassing, self).__init__()

        self.aggr = aggr
        assert self.aggr in ['add', 'mean', 'max']

        self.flow = flow
        assert self.flow in ['source_to_target', 'target_to_source']

        self.__message_args__ = getargspec(self.message)[0][1:]
        self.__special_args__ = [(i, arg)
                                 for i, arg in enumerate(self.__message_args__)
                                 if arg in special_args]
        self.__message_args__ = [
            arg for arg in self.__message_args__ if arg not in special_args
        ]
        self.__update_args__ = getargspec(self.update)[0][2:]

    def propagate(self, edge_index, size=None, dim=0, **kwargs):
        r"""The initial call to start propagating messages.
        Args:
            edge_index (Tensor): The indices of a general (sparse) assignment
                matrix with shape :obj:`[N, M]` (can be directed or
                undirected).
            size (list or tuple, optional): The size :obj:`[N, M]` of the
                assignment matrix. If set to :obj:`None`, the size is tried to
                get automatically inferred and assumed to be symmetric.
                (default: :obj:`None`)
            dim (int, optional): The axis along which to aggregate.
                (default: :obj:`0`)
            **kwargs: Any additional data which is needed to construct messages
                and to update node embeddings.
        """

        dim = 0
        size = [None, None] if size is None else list(size)
        assert len(size) == 2

        i, j = (0, 1) if self.flow == 'target_to_source' else (1, 0)
        ij = {"_i": i, "_j": j}

        message_args = []
        for arg in self.__message_args__:
            if arg[-2:] in ij.keys():
                tmp = kwargs.get(arg[:-2], None)
                if tmp is None:  # pragma: no cover
                    message_args.append(tmp)
                else:
                    idx = ij[arg[-2:]]
                    if isinstance(tmp, tuple) or isinstance(tmp, list):
                        assert len(tmp) == 2
                        if tmp[1 - idx] is not None:
                            if size[1 - idx] is None:
                                size[1 - idx] = tmp[1 - idx].size(dim)
                            if size[1 - idx] != tmp[1 - idx].size(dim):
                                raise ValueError(__size_error_msg__)
                        tmp = tmp[idx]

                    if tmp is None:
                        message_args.append(tmp)
                    else:
                        if size[idx] is None:
                            size[idx] = tmp.size(dim)
                        if size[idx] != tmp.size(dim):
                            raise ValueError(__size_error_msg__)

                        tmp = torch.index_select(tmp, dim, edge_index[idx])
                        message_args.append(tmp)
            else:
                message_args.append(kwargs.get(arg, None))

        size[0] = size[1] if size[0] is None else size[0]
        size[1] = size[0] if size[1] is None else size[1]

        kwargs['edge_index'] = edge_index
        kwargs['size'] = size

        for (idx, arg) in self.__special_args__:
            if arg[-2:] in ij.keys():
                message_args.insert(idx, kwargs[arg[:-2]][ij[arg[-2:]]])
            else:
                message_args.insert(idx, kwargs[arg])

        update_args = [kwargs[arg] for arg in self.__update_args__]

        out = self.message(*message_args)
        out = scatter_(self.aggr, out, edge_index[i], dim, dim_size=size[i])
        out = self.update(out, *update_args)

        return out

    def message(self, x_j):  # pragma: no cover
        r"""Constructs messages to node :math:`i` in analogy to
        :math:`\phi_{\mathbf{\Theta}}` for each edge in
        :math:`(j,i) \in \mathcal{E}` if :obj:`flow="source_to_target"` and
        :math:`(i,j) \in \mathcal{E}` if :obj:`flow="target_to_source"`.
        Can take any argument which was initially passed to :meth:`propagate`.
        In addition, tensors passed to :meth:`propagate` can be mapped to the
        respective nodes :math:`i` and :math:`j` by appending :obj:`_i` or
        :obj:`_j` to the variable name, *.e.g.* :obj:`x_i` and :obj:`x_j`.
        """

        return x_j

    def update(self, aggr_out):  # pragma: no cover
        r"""Updates node embeddings in analogy to
        :math:`\gamma_{\mathbf{\Theta}}` for each node
        :math:`i \in \mathcal{V}`.
        Takes in the output of aggregation as first argument and any argument
        which was initially passed to :meth:`propagate`."""

        return aggr_out


"""
auhor: ytlin
usage: convert csv to jsonl format
command: python csv2jsonl.py -i <csv-file-name> -o <jsonl-file-name>
caveat: the default seperator (delimiter) is '\t'
"""

import csv
import json
import argparse
import logging

"""
usage:
@MNLI/data/hans
python ../../src/data_git/csv2jsonl.py -i ./heuristics_evaluation_set.txt -o ./dev.jsonl
"""

logger = logging.getLogger(__name__)

if __name__ == "__main__":
    # parse arg
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_csv", "-i", type=str, required=True)
    parser.add_argument("--output_jsonl", "-o", type=str, required=True)
    args = parser.parse_args()
    # conversion
    with open(args.input_csv, 'r') as csvfile:
        with open(args.output_jsonl, 'w') as jsonfile:
            reader = csv.DictReader(csvfile, delimiter='\t') # use default fields
            for row in reader:
                json.dump(row, jsonfile)
                jsonfile.write('\n')
    # msg
    logger.info(msg=f"first three lines of jsonl {str(args.output_jsonl)}")
    with open(args.output_jsonl, 'r') as jsonfile:
        for i in range(3):
            example = jsonfile.readline()
            example = json.loads(example)
            logger.info(msg=f"{example}")
        
    logger.info(msg="csv to jsonl conversion done")


#!/usr/bin/env python
# set loggers here? see allenlp main
#import src.config

import os, sys

def run():
    print(__file__)
    print(os.path.dirname(os.path.realpath(__file__)))
    return


if __name__ == "__main__":
    run()


from typing import Union, Dict, List
from pathlib import Path

import torch 
import numpy as np
import matplotlib.pyplot as plt

def show_sequence_attention(seq: List[str], att: Union[torch.Tensor, np.array], serialization_dir: Union[str, Path]=None) -> None:
    """
    attention visualization
    seq is List[List[Token]]
    att is {"data": torch.Tensor, "mask": torch.BoolTensor}
    """
    # Set up figure with colorbar
    fig = plt.figure() # figsize=None
    ax = fig.add_subplot(111)
    # data = att["data"][0][att["mask"][0] ==True].T
    # if use raw batch att return
    # print(data.shape)
    cax = ax.imshow(att, cmap='bone', origin='upper')
    fig.colorbar(cax)
    # Set up axes
    ax.set_xticks(np.arange(len(seq)))
    ax.set_xticklabels(seq)
    ax.set_yticklabels([])
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")


    # Show label at every tick
    #ax.xaxis.set_major_locator(ticker.MultipleLocator(1))
    #ax.yaxis.set_major_locator(ticker.MultipleLocator(1))
    ax.set_title("Attention")
    fig.tight_layout()
    if serialization_dir is not None:
        plt.savefig(serialization_dir)
    plt.show()
    
def show_matrix_attention(seq1: List[str], seq2: List[str], att: Union[torch.Tensor, np.array], serialization_dir: Union[str, Path]=None)->None:
    """
    attention visualization
    seq1/seq2 is List[Token]
    att is 
    """
    # Set up figure with colorbar
    fig = plt.figure()
    ax = fig.add_subplot(111)
    cax = ax.matshow(att, cmap='bone')
    fig.colorbar(cax)

    # Set up axes
    ax.set_xticks(np.arange(len(seq1)))
    ax.set_yticks(np.arange(len(seq2)))
    ax.set_xticklabels(seq1, rotation=90)
    ax.set_yticklabels(seq2)

    # Show label at every tick
    #import matplotlib.ticker as ticker
    #ax.xaxis.set_major_locator(ticker.MultipleLocator(1))
    #ax.yaxis.set_major_locator(ticker.MultipleLocator(1))
    ax.set_title("Attention")
    fig.tight_layout()
    if serialization_dir is not None:
        plt.savefig(serialization_dir)
    plt.show()


# MODEL NAME for not config execution
TRANSFORMER_NAME = "bert-base-uncased"

# param for reader
# ANLI
anli_p_field = "context"
anli_h_field = "hypothesis"
anli_l_field = "label"
anli_fields = [anli_p_field, anli_h_field, anli_l_field]
# MNLI, SNLI
default_fields = ["sentence1", "sentence2", "gold_label"]

# label
label_to_id = {
    "contradiction" : 0,
    "neutral" : 1,
    "entailment" : 2,
}
id_to_label = ["contradiction", "neutral", "entailment"]


from overrides import overrides
from typing import Optional, Dict, Iterable, List, Union, Tuple

import torch_geometric
import torch

from allennlp.common import Registrable

from src.modules.graph2graph_encoders import  Graph2GraphEncoder
from src.modules.graph2vec_encoders import Graph2VecEncoder
from src.modules.graph_pair2graph_pair_encoders import GraphPair2GraphPairEncoder 
from src.modules.node_updaters import NodeUpdater

from src.modules.graph_pair2vec_encoders.graph_pair2vec_encoder import GraphPair2VecEncoder 

@GraphPair2VecEncoder.register("graph_matching_net", exist_ok=True)
class GraphMatchingNet(GraphPair2VecEncoder):
    """
    `GraphMatchingNet` differs from `GraphEmbeddingNet` with an extra cross graph attention.
    
    In each layer,
    `Graph2GraphEncoder` encodes each graph seperately,
    `GraphMatchingNet` matches two graph by a similarity function and produces the soft-aligned version of the other graph for each graph,
    `NodeUpdater` than is used to update the current node repr by the msgs.
    
    After L layers of encoding and matching,'Graph2VecEncoder' is applied to project 2 graphs into the same representation space,
    then return a vector $[g1;g2;g1-g2;g1 \odot g2]$ for further classification
    """
    
    def __init__(
        self,
        num_layers: int,
        convs: Graph2GraphEncoder, 
        atts: GraphPair2GraphPairEncoder,
        updaters: List[NodeUpdater],  
        poolers: List[Graph2VecEncoder], 
    ) -> None:
        """
        `GraphMatchingNet` constructor
        """
        super(GraphMatchingNet, self).__init__()
        # if given is not List[item], create List[item]
        if not isinstance(convs, list):
            convs = [convs] * num_layers # share param
        if not isinstance(atts, list):
            atts = [atts] * num_layers
            
        if len(convs) != num_layers:
            raise ConfigurationError(
                "len(convs) (%d) != num_layers (%d)" % (len(convs), num_layers)
            )
        if len(atts) != num_layers:
            raise ConfigurationError(
                "len(atts) (%d) != num_layers (%d)" % (len(atts), num_layers)
            )
        
            
        self._convs = torch.nn.ModuleList(convs)
        self._atts = torch.nn.ModuleList(atts)
        self._updaters = torch.nn.ModuleList(updaters)
        self._output_dim = 4*convs[-1].get_output_dim() # for vector pair comparison 
        self._input_dim = convs[0].get_input_dim()
        self._poolers = torch.nn.ModuleList(poolers)
        self.num_layers = num_layers
    
    @overrides
    def get_output_dim(self):
        return self._output_dim
    
    @overrides
    def get_input_dim(self):
        return self._input_dim

    def forward(
        self,
        x1: Dict[str, torch.Tensor],
        x2: Dict[str, torch.Tensor],
        g1: Dict[str, torch.Tensor],
        g2: Dict[str, torch.Tensor],
        return_attention: bool = False,
    ): # -> Union[torch.Tensor, Tuple]:
        """
        input:
            g1, g2 : Dict[str, torch.Tensor] sparse_adjacency batch
            n1, n2 : Dict[str, torch.Tensor] sparse_node_embedding batch
            e1, e2 : OptTensor sparse_edge_embedding batch
            return_attention : bool, flag to turn on or off attention return
        """
        # node_tensor, node_batch_indices
        x1, b1 = x1["data"], x1["batch_indices"]
        x2, b2 = x2["data"], x2["batch_indices"]
        # edge_index, edge_type, edge_batch
        e1, t1, eb1 = g1['edge_index'], g1['edge_attr'], g1['batch_id']
        e2, t2, eb2 = g2['edge_index'], g2['edge_attr'], g2['batch_id']
        # return attention
        if return_attention:
            atts = {}
        
        # apply Graph2Graph Encoders by module list
        for i in range(self.num_layers):
            # calculate message (n_nodes, dim_encoder)
            from src.modules.graph2graph_encoders.HGTConv import HGTConv
            if isinstance(self._convs[i], HGTConv):
                #print("debug")
                x1_msg = self._convs[i](x=x1, edge_index=e1, edge_type=t1)
                x2_msg = self._convs[i](x=x2, edge_index=e2, edge_type=t2)
            else:
                x1_msg = self._convs[i](x=x1, edge_index=e1, edge_type=t1)
                x2_msg = self._convs[i](x=x2, edge_index=e2, edge_type=t2)
            # calculate matching (n_nodes, dim_matching)
            if return_attention:
                x1_match, x2_match, matching_attention = self._atts[i](x1, x2, b1, b2, return_attention)
                atts[f"matching{i}"] = matching_attention
            else:
                x1_match, x2_match = self._atts[i](x1, x2, b1, b2)
            # update (n_nodes, dim_encoder)
            x1 = self._updaters[0]([x1_msg, x1_match], x1)
            x2 = self._updaters[1]([x2_msg, x2_match], x2)
        
        # Graph Pooling => (batch_size, _input_dim)
        if return_attention:
            v1, pooler_att1 = self._poolers[0](x1, batch=b1, return_attention=return_attention)
            v2, pooler_att2 = self._poolers[1](x2, batch=b2, return_attention=return_attention)
            atts["pooler1"] = pooler_att1
            atts["pooler2"] = pooler_att2
        else:
            v1 = self._poolers[0](x1, batch=b1, return_attention=return_attention)
            v2 = self._poolers[1](x2, batch=b2, return_attention=return_attention)
        # Shape: (batch_size, _out_put_dim)
        out = torch.cat([v1, v2, v1-v2, v1*v2], dim=1)
        
        if return_attention:
            return out, atts
        else:
            return out



from overrides import overrides
from typing import Optional, Dict, Iterable, List, Union

#import torch_geometric
import torch

from allennlp.common import Registrable

class GraphPair2VecEncoder(torch.nn.Module, Registrable):
    """
    A `GraphPair2VecEncoder` is a `Module` that takes
    two sequence of vectors and two graphs
    and returns
    a single vector.
    
    Input shape : `(SparseBatch, SparseGraphBatch)`;
    output shape: `(batch_size, output_dimension)`.
    """
    def get_input_dim(self) -> int:
        """
        Returns the dimension of the vector input for each element in the sequence input
        to a `Seq2VecEncoder`.
        """
        raise NotImplementedError
        
    def get_output_dim(self) -> int:
        """
        Returns the dimension of the final vector output by this `Seq2VecEncoder`.  This is `not`
        the shape of the returned tensor, but the last element of that shape.
        """
        raise NotImplementedError




from overrides import overrides
from typing import Optional, Dict, Iterable, List, Union

import torch_geometric
import torch

from allennlp.common import Registrable

from src.modules.graph2graph_encoders import  Graph2GraphEncoder
from src.modules.graph2vec_encoders import Graph2VecEncoder
from src.modules.graph_pair2graph_pair_encoders import GraphPair2GraphPairEncoder 

from src.modules.graph_pair2vec_encoders.graph_pair2vec_encoder import GraphPair2VecEncoder 

@GraphPair2VecEncoder.register("graph_embedding_net", exist_ok=True)
class GraphEmbeddingNet(GraphPair2VecEncoder):
    """
    `GraphEmbeddingNet` encodes 2 graphs with `Graph2GraphEncoder` seperately,
    then use 'Graph2VecEncoder' to project 2 graphs into the same representation space,
    then return a vector $[g1;g2;g1-g2;g1 \odot g2]$ for further classification
    """
    
    """ todo: will uncomment after test
    __slots__ = (
        "_input_dim",
        "_output_dim",
        "_convs",
        "num_layers"
    )
    """
    
    def __init__(
        self,
        num_layers: int,
        convs: Graph2GraphEncoder, # todo: list version
        pooler: Graph2VecEncoder, # graph pooler for mapping graph to embedding space
        #**kwargs, todo: will, 
        #dropout: Union[float, List[float]] = 0.0,
    ) -> None:
        """
        Old Style(I think this is not ok since if convname is not the same, they may need different params)
        `GraphEmbeddingNet` constructor
        note that convs is str or List[str] (in Graph2GraphEncoder.list_availabels() )
        so that this constructor calls Graph2GraphEncoder.from_name(convs[i]) for constructing convolutions
        """
        super(GraphEmbeddingNet, self).__init__()
        # if given is not List[item], create List[item]
        # this method implicitly share params? 
        if not isinstance(convs, list):
            convs = [convs] * num_layers
            
        if len(convs) != num_layers:
            raise ConfigurationError(
                "len(convs) (%d) != num_layers (%d)" % (len(convs), num_layers)
            )
            
        self._convs = torch.nn.ModuleList(convs)
        self._output_dim = 4*convs[-1].get_output_dim() # for vector pair comparison 
        self._input_dim = convs[0].get_input_dim()
        self._pooler = pooler
        self.num_layers = num_layers
    
    @overrides
    def get_output_dim(self):
        return self._output_dim
    
    @overrides
    def get_input_dim(self):
        return self._input_dim

    def forward(
        self,
        x1: Dict[str, torch.Tensor],
        x2: Dict[str, torch.Tensor],
        g1: Dict[str, torch.Tensor],
        g2: Dict[str, torch.Tensor],
        return_attention: bool=False,
    ) -> torch.Tensor:
        """
        input:
            g1, g2 : Dict[str, torch.Tensor] sparse_adjacency batch
            n1, n2 : Dict[str, torch.Tensor] sparse_node_embedding batch
            e1, e2 : OptTensor sparse_edge_embedding batch
        """
        # node_tensor, node_batch_indices
        x1, b1 = x1["data"], x1["batch_indices"]
        x2, b2 = x2["data"], x2["batch_indices"]
        # edge_index, edge_type, edge_batch
        e1, t1, eb1 = g1['edge_index'], g1['edge_attr'], g1['batch_id']
        e2, t2, eb2 = g2['edge_index'], g2['edge_attr'], g2['batch_id']
        # return attention
        if return_attention:
            atts = {}
        
        # Graph Pooling => (batch_size, _input_dim)
        if return_attention:
            v1, pooler_att1 = self._pooler(x1, batch=b1, return_attention=return_attention)
            v2, pooler_att2 = self._pooler(x2, batch=b2, return_attention=return_attention)
            atts["pooler1"] = pooler_att1
            atts["pooler2"] = pooler_att2
        else:
            v1 = self._pooler(x1, batch=b1, return_attention=return_attention)
            v2 = self._pooler(x2, batch=b2, return_attention=return_attention)
        # Shape: (batch_size, _out_put_dim)
        out = torch.cat([v1, v2, v1-v2, v1*v2], dim=1)
        
        if return_attention:
            return out, atts
        else:
            return out



from typing import Dict, List, Set, Tuple, Union, TypeVar
import logging
import textwrap

from collections import defaultdict
from overrides import overrides
import torch

from allennlp.common.checks import ConfigurationError
from allennlp.data.fields import (Field, TextField, SequenceField)
#from allennlp.data.fields.sequence_field import SequenceField
from allennlp.data.vocabulary import Vocabulary

from torch_geometric.data.data import Data as PyGeoData
from torch_geometric.data.batch import Batch as PyGeoBatch

DataArray = TypeVar(
    "DataArray", torch.Tensor, Dict[str, torch.Tensor], Dict[str, Dict[str, torch.Tensor]]
)

logger = logging.getLogger(__name__)

# for input model
SparseAdjacencyFieldTensors = Dict[str, torch.Tensor]

# author = ytlin 
class SparseAdjacencyField(Field[torch.Tensor]):
    """
    A `SparseAdjacencyField` defines directed adjacency relations between elements
    in a :class:`~allennlp.data.fields.sequence_field.SequenceField`.
    Because it's a labeling of some other field, we take that field as input here
    and use it to determine our padding and other things.
    This field will handle the batching and vocaburary indexing of pytorch geometric Data type.
    
    # Parameters
    graph: `PyGeoData`
        Note that in PyGeoData
        edge_attr is list of edge attributes (List[Any])
        node_attr is list of node attributes (List[Any])
        edge_index is a [2, |V|] tensor
        # can be a way to implement this w/o using PyGeoData
    sequence_field: `SequenceField`
        I do not construct sequencefield from node_attr, since need tokenizer 
    label_namespace : `str`, optional (default=`'labels'`)
        The namespace to use for converting tag strings into integers.  We convert tag strings to
        integers for you, and this parameter tells the `Vocabulary` object which mapping from
        strings to integers to use (so that "O" as a tag doesn't get the same id as "O" as a word).
    #  do not need padding here, may need follow_batch (but in this implementation, we treat node_attr to follow_batch)
    
    """
    __slots__ = (
        "edge_indices",
        "labels",
        "sequence_field",
        "_label_namespace",
        "_indexed_labels",
    )
    

    # It is possible that users want to use this field with a namespace which uses OOV/PAD tokens.
    # This warning will be repeated for every instantiation of this class (i.e for every data
    # instance), spewing a lot of warnings so this class variable is used to only log a single
    # warning per namespace.
    _already_warned_namespaces: Set[str] = set()

    def __init__(
        self,
        graph: PyGeoData,
        sequence_field: SequenceField,
        label_namespace: str = "edge_labels"
    ) -> None:
        # Field inheritor dose not need to super().__init__() 
        # label skip index is not implemented yet, todo
        labels: Union[List[str], List[int]] = graph.edge_attr
        self.edge_indices = graph.edge_index
        self.labels = labels
        self.sequence_field = sequence_field
        self._label_namespace = label_namespace
        self._indexed_labels: List[int] = None

        self._maybe_warn_for_namespace(label_namespace)
        field_length = sequence_field.sequence_length()
        
        # we do not check duplicate edge, since edge can be multi label
        #if len(set(edge_indices)) != len(edge_indices):
        #    raise ConfigurationError(f"edge_indices must be unique, but found {edge_indices}")
        
        if self.edge_indices is not None: # do not check for empty_field
            # check for out-of-index edge
            if not all(
                0 <= self.edge_indices[1][i] < field_length and 0 <= self.edge_indices[0][i] < field_length for i in range(self.edge_indices.size()[1])
            ):
                raise ConfigurationError(
                    f"Label edge_indices and sequence length "
                    f"are incompatible: {self.edge_indices} and {field_length}"
                )

            # if labels is passed, should have same length with edges 
            if labels is not None and self.edge_indices.size()[1] != len(labels):
                raise ConfigurationError(
                    f"Labelled edge_indices were passed, but their lengths do not match: "
                    f" {labels}, {self.edge_indices}"
                )

    def _maybe_warn_for_namespace(self, label_namespace: str) -> None:
        if not (self._label_namespace.endswith("labels") or self._label_namespace.endswith("tags")):
            if label_namespace not in self._already_warned_namespaces:
                logger.warning(
                    "Your label namespace was '%s'. We recommend you use a namespace "
                    "ending with 'labels' or 'tags', so we don't add UNK and PAD tokens by "
                    "default to your vocabulary.  See documentation for "
                    "`non_padded_namespaces` parameter in Vocabulary.",
                    self._label_namespace,
                )
                self._already_warned_namespaces.add(label_namespace)

    @overrides
    def count_vocab_items(self, counter: Dict[str, Dict[str, int]]):
        if self._indexed_labels is None and self.labels is not None:
            for label in self.labels:
                counter[self._label_namespace][label] += 1  # type: ignore

    @overrides
    def index(self, vocab: Vocabulary):
        if self.labels is not None:
            self._indexed_labels = [
                vocab.get_token_index(label, self._label_namespace) for label in self.labels
            ]
    
    @overrides
    def get_padding_lengths(self) -> Dict[str, int]:
        """
        If there are things in this field that need padding, note them here. 
        In order to pad a batch of instance, we get all of the lengths from the batch, take the max, and pad
        everything to that length (or use a pre-specified maximum length).
        The return value is a dictionary mapping keys to lengths, like `{'num_tokens': 13}`.
        This is always called after :func:`index`.
        """
        return {"num_tokens": len(self.sequence_field)}
    
    # this is the main difference between Sparse and Dense Adjacency
    @overrides
    def as_tensor(self, padding_lengths: Dict[str, int]) -> torch.Tensor:
        tensor_dict = {}
        tensor_dict["edge_attr"] = torch.tensor(self._indexed_labels, dtype=torch.long)
        tensor_dict["edge_index"] = self.edge_indices
        tensor_dict["batch_id"] = torch.zeros(len(self.sequence_field)) # batch_id for the node_attr
        return tensor_dict
    
    # use pytorch geometric sparse graph batching style
    # I looked up the references for coding this part
    # https://pytorch-geometric.readthedocs.io/en/latest/_modules/torch_geometric/data/batch.html#Batch.from_data_list
    # https://github.com/allenai/allennlp/blob/master/allennlp/data/fields/text_field.py
    @overrides 
    def batch_tensors(self, tensor_list: List[DataArray]) -> DataArray:  # type: ignore
        """
        Takes the output of `Field.as_tensor()` from a list of `Instances` and merges it into
        one batched tensor for this `Field`. 
        
        Input: List[TensorDict]
        Output: TensorDict
        """
        pre_sum = 0
        batch_tensor = defaultdict(list)
        
        for batch_id, tensor in enumerate(tensor_list):
            cur_edge_index = tensor["edge_index"].add(pre_sum)
            cur_nodes = len(tensor["batch_id"])
            batch_tensor["edge_index"].append(cur_edge_index)
            batch_tensor["edge_attr"].append(tensor["edge_attr"])
            cur_id_tensor = torch.full(size=[cur_nodes], fill_value=batch_id, dtype=torch.long) # as a 1*d tensor
            batch_tensor["batch_id"].append(cur_id_tensor)
            pre_sum += cur_nodes
        batch_tensor["edge_index"] = torch.cat(batch_tensor["edge_index"], dim=1)
        batch_tensor["edge_attr"] = torch.cat(batch_tensor["edge_attr"], dim=0)
        batch_tensor["batch_id"] = torch.cat(batch_tensor["batch_id"], dim=0)
        return dict(batch_tensor)
    
    # tested that can build, not know whats for, may be errors in downstream
    @overrides
    def empty_field(self) -> "SparseAdjacencyField":

        # The empty_list here is needed for mypy
        empty_adjacency_field = SparseAdjacencyField(
            graph=PyGeoData(), sequence_field=self.sequence_field.empty_field()
        )
        return empty_adjacency_field

    def __str__(self) -> str:
        length = self.sequence_field.sequence_length()
        formatted_labels = "".join(
            "\t\t" + labels + "\n" for labels in textwrap.wrap(repr(self.labels), 100)
        )
        formatted_edge_indices = "".join(
            "\t\t" + index + "\n" for index in textwrap.wrap(repr(self.edge_indices), 100)
        )
        return (
            f"AdjacencyField of length {length}\n"
            f"\t\twith edge_indices:\n {formatted_edge_indices}\n"
            f"\t\tand labels:\n {formatted_labels} \t\tin namespace: '{self._label_namespace}'."
        )

    def __len__(self):
        return len(self.sequence_field)


"""
Caveat: do not use this now
I will need to rewrite my own

This part is from `PyHGT`
modified part :
1. remove **kwargs
2. assert(out_dim % n_heads == 0)
3. change forward
4. add get_output_dim()
5. MessagePassing the old stype(1.3.2) is used... (Caveat)

usage of mine:

about node type:
can use POS tags or ignore(since relation should have this info?)

about edge type:
exactly what we need

about edge time:
depdency graph is not graph with temporal difference, so pass all with the same num is enough

the math can be found in `https://arxiv.org/pdf/2003.01332.pdf`

"type": "hgt"
"in_dim" : 300
"out_dim" : 300
"num_types" : 10
"num_relations" : 20
"n_heads" : 5
"dropout" : 0.2
"use_norm" : true
"use_RTE" : false

"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
#from torch_geometric.nn.conv import MessagePassing
from .MessagePassing132 import MessagePassing
from torch_geometric.nn.inits import glorot, uniform
from torch_geometric.utils import softmax
import math

class HGTConv(MessagePassing):
    def __init__(self, in_dim, out_dim, num_types, num_relations, n_heads, dropout = 0.2, use_norm = True, use_RTE = True):
        super(HGTConv, self).__init__(aggr='add')

        self.in_dim        = in_dim
        self.out_dim       = out_dim
        self.num_types     = num_types
        self.num_relations = num_relations
        self.total_rel     = num_types * num_relations * num_types
        self.n_heads       = n_heads
        assert(out_dim % n_heads == 0)
        self.d_k           = out_dim // n_heads
        self.sqrt_dk       = math.sqrt(self.d_k)
        self.use_norm      = use_norm
        self.use_RTE       = use_RTE
        self.att           = None
        
        
        self.k_linears   = nn.ModuleList()
        self.q_linears   = nn.ModuleList()
        self.v_linears   = nn.ModuleList()
        self.a_linears   = nn.ModuleList()
        self.norms       = nn.ModuleList()
        
        for t in range(num_types):
            self.k_linears.append(nn.Linear(in_dim,   out_dim))
            self.q_linears.append(nn.Linear(in_dim,   out_dim))
            self.v_linears.append(nn.Linear(in_dim,   out_dim))
            self.a_linears.append(nn.Linear(out_dim,  out_dim))
            if use_norm:
                self.norms.append(nn.LayerNorm(out_dim))
        '''
            TODO: make relation_pri smaller, as not all <st, rt, tt> pair exist in meta relation list.
        '''
        self.relation_pri   = nn.Parameter(torch.ones(num_relations, self.n_heads))
        self.relation_att   = nn.Parameter(torch.Tensor(num_relations, n_heads, self.d_k, self.d_k))
        self.relation_msg   = nn.Parameter(torch.Tensor(num_relations, n_heads, self.d_k, self.d_k))
        self.skip           = nn.Parameter(torch.ones(num_types))
        self.drop           = nn.Dropout(dropout)
        
        if self.use_RTE:
            self.emb            = RelTemporalEncoding(in_dim)
        
        # reset parameters
        glorot(self.relation_att)
        glorot(self.relation_msg)
    
    def get_input_dim(self):
        return self.in_dim
    
    def get_output_dim(self):
        return self.out_dim
        
    def forward(self, x, edge_index, edge_type, edge_time=None, node_type=None):
        # to make compatible with graph matching net
        if node_type is None:
            node_type = x.new_zeros(x.size()[0]) # .unsqueeze(1)
            #print(node_type.size())
        if edge_time is None:
            edge_time = edge_type.new_zeros(edge_type.size()) #.unsqueeze(1)
            #print(edge_time.size())
        return self.propagate(edge_index, node_inp=x, node_type=node_type, \
                              edge_type=edge_type, edge_time = edge_time)

    def message(self, edge_index_i, node_inp_i, node_inp_j, node_type_i, node_type_j, edge_type, edge_time):
        '''
            j: source, i: target; <j, i>
        '''
        data_size = edge_index_i.size(0)
        '''
            Create Attention and Message tensor beforehand.
        '''
        res_att     = torch.zeros(data_size, self.n_heads).to(node_inp_i.device)
        res_msg     = torch.zeros(data_size, self.n_heads, self.d_k).to(node_inp_i.device)
        
        for source_type in range(self.num_types):
            sb = (node_type_j == int(source_type))
            k_linear = self.k_linears[source_type]
            v_linear = self.v_linears[source_type] 
            for target_type in range(self.num_types):
                tb = (node_type_i == int(target_type)) & sb
                q_linear = self.q_linears[target_type]
                for relation_type in range(self.num_relations):
                    '''
                        idx is all the edges with meta relation <source_type, relation_type, target_type>
                    '''
                    idx = (edge_type == int(relation_type)) & tb
                    if idx.sum() == 0:
                        continue
                    '''
                        Get the corresponding input node representations by idx.
                        Add tempotal encoding to source representation (j)
                    '''
                    target_node_vec = node_inp_i[idx]
                    source_node_vec = node_inp_j[idx]
                    if self.use_RTE:
                        source_node_vec = self.emb(source_node_vec, edge_time[idx])
                    '''
                        Step 1: Heterogeneous Mutual Attention
                    '''
                    q_mat = q_linear(target_node_vec).view(-1, self.n_heads, self.d_k)
                    k_mat = k_linear(source_node_vec).view(-1, self.n_heads, self.d_k)
                    k_mat = torch.bmm(k_mat.transpose(1,0), self.relation_att[relation_type]).transpose(1,0)
                    res_att[idx] = (q_mat * k_mat).sum(dim=-1) * self.relation_pri[relation_type] / self.sqrt_dk
                    '''
                        Step 2: Heterogeneous Message Passing
                    '''
                    v_mat = v_linear(source_node_vec).view(-1, self.n_heads, self.d_k)
                    res_msg[idx] = torch.bmm(v_mat.transpose(1,0), self.relation_msg[relation_type]).transpose(1,0)   
        '''
            Softmax based on target node's id (edge_index_i). Store attention value in self.att for later visualization.
        '''
        self.att = softmax(res_att, edge_index_i)
        res = res_msg * self.att.view(-1, self.n_heads, 1)
        del res_att, res_msg
        return res.view(-1, self.out_dim)


    def update(self, aggr_out, node_inp, node_type):
        '''
            Step 3: Target-specific Aggregation
            x = W[node_type] * gelu(Agg(x)) + x
        '''
        aggr_out = F.gelu(aggr_out)
        res = torch.zeros(aggr_out.size(0), self.out_dim).to(node_inp.device)
        for target_type in range(self.num_types):
            idx = (node_type == int(target_type))
            if idx.sum() == 0:
                continue
            trans_out = self.drop(self.a_linears[target_type](aggr_out[idx]))
            '''
                Add skip connection with learnable weight self.skip[t_id]
            '''
            alpha = torch.sigmoid(self.skip[target_type])
            if self.use_norm:
                res[idx] = self.norms[target_type](trans_out * alpha + node_inp[idx] * (1 - alpha))
            else:
                res[idx] = trans_out * alpha + node_inp[idx] * (1 - alpha)
        return res

    def __repr__(self):
        return '{}(in_dim={}, out_dim={}, num_types={}, num_types={})'.format(
            self.__class__.__name__, self.in_dim, self.out_dim,
            self.num_types, self.num_relations)
    
    
    
class DenseHGTConv(MessagePassing):
    def __init__(self, in_dim, out_dim, num_types, num_relations, n_heads, dropout = 0.2, use_norm = True, use_RTE = True):
        super(DenseHGTConv, self).__init__(aggr='add')

        self.in_dim        = in_dim
        self.out_dim       = out_dim
        self.num_types     = num_types
        self.num_relations = num_relations
        self.total_rel     = num_types * num_relations * num_types
        self.n_heads       = n_heads
        self.d_k           = out_dim // n_heads
        self.sqrt_dk       = math.sqrt(self.d_k)
        self.use_norm      = use_norm
        self.use_RTE       = use_RTE
        self.att           = None
        
        
        self.k_linears   = nn.ModuleList()
        self.q_linears   = nn.ModuleList()
        self.v_linears   = nn.ModuleList()
        self.a_linears   = nn.ModuleList()
        self.norms       = nn.ModuleList()

        
        for t in range(num_types):
            self.k_linears.append(nn.Linear(in_dim,   out_dim))
            self.q_linears.append(nn.Linear(in_dim,   out_dim))
            self.v_linears.append(nn.Linear(in_dim,   out_dim))
            self.a_linears.append(nn.Linear(out_dim,  out_dim))
            if use_norm:
                self.norms.append(nn.LayerNorm(out_dim))
        '''
            TODO: make relation_pri smaller, as not all <st, rt, tt> pair exist in meta relation list.
        '''
        self.relation_pri   = nn.Parameter(torch.ones(num_relations, self.n_heads))
        self.relation_att   = nn.Parameter(torch.Tensor(num_relations, n_heads, self.d_k, self.d_k))
        self.relation_msg   = nn.Parameter(torch.Tensor(num_relations, n_heads, self.d_k, self.d_k))
        self.drop           = nn.Dropout(dropout)
        
        if self.use_RTE:
            self.emb            = RelTemporalEncoding(in_dim)
        
        glorot(self.relation_att)
        glorot(self.relation_msg)
        
        
        self.mid_linear  = nn.Linear(out_dim,  out_dim * 2)
        self.out_linear  = nn.Linear(out_dim * 2,  out_dim)
        self.out_norm    = nn.LayerNorm(out_dim)
        
    def forward(self, node_inp, node_type, edge_index, edge_type, edge_time):
        return self.propagate(edge_index, node_inp=node_inp, node_type=node_type, \
                              edge_type=edge_type, edge_time = edge_time)

    def message(self, edge_index_i, node_inp_i, node_inp_j, node_type_i, node_type_j, edge_type, edge_time):
        '''
            j: source, i: target; <j, i>
        '''
        data_size = edge_index_i.size(0)
        '''
            Create Attention and Message tensor beforehand.
        '''
        res_att     = torch.zeros(data_size, self.n_heads).to(node_inp_i.device)
        res_msg     = torch.zeros(data_size, self.n_heads, self.d_k).to(node_inp_i.device)
        
        for source_type in range(self.num_types):
            sb = (node_type_j == int(source_type))
            k_linear = self.k_linears[source_type]
            v_linear = self.v_linears[source_type] 
            for target_type in range(self.num_types):
                tb = (node_type_i == int(target_type)) & sb
                q_linear = self.q_linears[target_type]
                for relation_type in range(self.num_relations):
                    '''
                        idx is all the edges with meta relation <source_type, relation_type, target_type>
                    '''
                    idx = (edge_type == int(relation_type)) & tb
                    if idx.sum() == 0:
                        continue
                    '''
                        Get the corresponding input node representations by idx.
                        Add tempotal encoding to source representation (j)
                    '''
                    target_node_vec = node_inp_i[idx]
                    source_node_vec = node_inp_j[idx]
                    if self.use_RTE:
                        source_node_vec = self.emb(source_node_vec, edge_time[idx])
                    '''
                        Step 1: Heterogeneous Mutual Attention
                    '''
                    q_mat = q_linear(target_node_vec).view(-1, self.n_heads, self.d_k)
                    k_mat = k_linear(source_node_vec).view(-1, self.n_heads, self.d_k)
                    k_mat = torch.bmm(k_mat.transpose(1,0), self.relation_att[relation_type]).transpose(1,0)
                    res_att[idx] = (q_mat * k_mat).sum(dim=-1) * self.relation_pri[relation_type] / self.sqrt_dk
                    '''
                        Step 2: Heterogeneous Message Passing
                    '''
                    v_mat = v_linear(source_node_vec).view(-1, self.n_heads, self.d_k)
                    res_msg[idx] = torch.bmm(v_mat.transpose(1,0), self.relation_msg[relation_type]).transpose(1,0)   
        '''
            Softmax based on target node's id (edge_index_i). Store attention value in self.att for later visualization.
        '''
        self.att = softmax(res_att, edge_index_i)
        res = res_msg * self.att.view(-1, self.n_heads, 1)
        del res_att, res_msg
        return res.view(-1, self.out_dim)


    def update(self, aggr_out, node_inp, node_type):
        '''
            Step 3: Target-specific Aggregation
            x = W[node_type] * Agg(x) + x
        '''
        res = torch.zeros(aggr_out.size(0), self.out_dim).to(node_inp.device)
        for target_type in range(self.num_types):
            idx = (node_type == int(target_type))
            if idx.sum() == 0:
                continue
            trans_out = self.drop(self.a_linears[target_type](aggr_out[idx])) + node_inp[idx]
            '''
                Add skip connection with learnable weight self.skip[t_id]
            '''
            if self.use_norm:
                trans_out = self.norms[target_type](trans_out)
                
            '''
                Step 4: Shared Dense Layer
                x = Out_L(gelu(Mid_L(x))) + x
            '''
                
            trans_out     = self.drop(self.out_linear(F.gelu(self.mid_linear(trans_out)))) + trans_out
            res[idx]      = self.out_norm(trans_out)
        return res

    def __repr__(self):
        return '{}(in_dim={}, out_dim={}, num_types={}, num_types={})'.format(
            self.__class__.__name__, self.in_dim, self.out_dim,
            self.num_types, self.num_relations)


class RelTemporalEncoding(nn.Module):
    '''
        Implement the Temporal Encoding (Sinusoid) function.
    '''
    def __init__(self, n_hid, max_len = 240, dropout = 0.2):
        super(RelTemporalEncoding, self).__init__()
        position = torch.arange(0., max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, n_hid, 2) *
                             -(math.log(10000.0) / n_hid))
        emb = nn.Embedding(max_len, n_hid)
        emb.weight.data[:, 0::2] = torch.sin(position * div_term) / math.sqrt(n_hid)
        emb.weight.data[:, 1::2] = torch.cos(position * div_term) / math.sqrt(n_hid)
        emb.requires_grad = False
        self.emb = emb
        self.lin = nn.Linear(n_hid, n_hid)
    def forward(self, x, t):
        return x + self.lin(self.emb(t))


from setuptools import setup, find_packages

# build long description from README.md
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="syn_nli", # Replace with your own username
    version="0.0.1",
    author="ytlin",
    author_email="0312fs3@gmail.com",
    description="package for the paper Syntax Aware Natural Language Inference@<link>",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EazyReal/2020-IIS-internship",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="allennlp NLP deep learning machine",
    license="Apache",
    packages=find_packages(
        exclude=[
            "*.tests",
            "*.tests.*",
            "tests.*",
            "tests",
            "test_fixtures",
            "test_fixtures.*",
            "benchmarks",
            "benchmarks.*",
            "previous_srcs"
        ]
    ),
    install_requires=[
        "torch>=1.6.0,<1.7.0",
        "allennlp==1.1.0rc3",
        "pytorch_geometic" # change this later
    ],
    entry_points={"console_scripts": ["src=src.__main__:run"]},
    include_package_data=True,
    python_requires=">=3.6.1",
    zip_safe=False,
)


"""
author = YT Lin
graph_nli.py is for the models that requires edge embedding,
can use Passthrough provided in allennlp if not required
"""
# external
import json
from typing import Dict, Iterable, List
from copy import deepcopy

# allennlp
import numpy as np
import torch

from allennlp.data import DatasetReader, DataLoader, Instance, Vocabulary
from allennlp.data.fields import LabelField, TextField
from allennlp.data.fields.text_field import TextFieldTensors
from allennlp.data.token_indexers import TokenIndexer, SingleIdTokenIndexer
from allennlp.data.tokenizers import Token, Tokenizer, WhitespaceTokenizer
from allennlp.models import Model
from allennlp.modules import TextFieldEmbedder, Seq2VecEncoder, FeedForward
#from allennlp.modules.seq2vec_encoders import BagOfEmbeddingsEncoder
from allennlp.nn import util
from allennlp.training.metrics import CategoricalAccuracy, Entropy

from allennlp.modules import TextFieldEmbedder, TokenEmbedder
from allennlp.modules.token_embedders import PretrainedTransformerMismatchedEmbedder
# BasicTextFieldEmbedder is Dict[str, TokenEmbedder]

from allennlp.nn import (
    InitializerApplicator, 
    RegularizerApplicator,
)

# FeedForward(124, 2, [64, 32], torch.nn.ReLU(), 0.2)

#self import 
# import src.config as config
from src.data_git.sparse_adjacency_field import SparseAdjacencyField, SparseAdjacencyFieldTensors
# for batch transform
import src.tensor_op as tensor_op
from src.modules.graph_pair2vec_encoders import GraphPair2VecEncoder

@Model.register("graph-nli-ee", exist_ok=True)
class GraphNLIModel(Model):
    def __init__(self,
                 vocab: Vocabulary,
                 embedder: TextFieldEmbedder,
                 edge_embedder: TokenEmbedder,
                 projector: FeedForward,
                 encoder: GraphPair2VecEncoder,
                 classifier: FeedForward,
                 initializer: InitializerApplicator = InitializerApplicator()
                ):
        """
        vocab : for edge_labels mainly
        embedder: text_token_ids => text_embedding_space
        gmn : GraphMatchingNetwork, take tokens, graph_adj pair to produce a single vector for cls
        cls : classifier
        """
        super().__init__(vocab)
        num_labels = vocab.get_vocab_size("labels") #3
        num_relations = vocab.get_vocab_size("relations") #20?
        self.embedder = embedder 
        self.edge_embedder = edge_embedder
        self.projector = projector
        self.encoder = encoder
        self.classifier = classifier
        self.accuracy = CategoricalAccuracy()
        self.entropy = Entropy()
        initializer(self) # init parameters, would not modify model slots
        # check dimension match if required
        return
        
    def forward(self,
            tokens_p: TextFieldTensors,
            tokens_h: TextFieldTensors,
            g_p: SparseAdjacencyFieldTensors,
            g_h: SparseAdjacencyFieldTensors,
            label: torch.Tensor = None,
            return_attention: bool = False,
        ) -> Dict[str, torch.Tensor]:
        """
        GMN for NLI
        let B be batch size
        let N be p length
        let M be h length
        
        input :
            tokens_p in shape [*] (B, N)
            g_p["edge_index"]
        ouput : tensor dict
        """
        # Shape: (batch_size, num_tokens, embedding_dim)
        # embedder will take out the desired entry, for ex: ["tokens"]
        embedded_p = self.embedder(tokens_p) 
        embedded_h = self.embedder(tokens_h)
        # Shape: (num_batch_edges, edge_embedding_dim)
        # can use pass through if type is the information required
        # this is general for any kind of edge representation to GP2Vencoder
        # https://docs.allennlp.org/master/api/modules/token_embedders/embedding/
        # inplace will change input!!!
        g_p_embedded = deepcopy(g_p)
        g_h_embedded = deepcopy(g_h)
        g_p_embedded["edge_attr"] = self.edge_embedder(g_p["edge_attr"])
        g_h_embedded["edge_attr"] = self.edge_embedder(g_h["edge_attr"])
        assert(not torch.any(torch.isnan(embedded_p)))
        assert(not torch.any(torch.isnan(embedded_h)))
        # Shape: (batch_size, num_tokens, projected_dim)
        embedded_p = self.projector(embedded_p)
        embedded_h = self.projector(embedded_h)
        assert(not torch.any(torch.isnan(embedded_p)))
        assert(not torch.any(torch.isnan(embedded_h)))
        # Shape:
        # node_attr : (num_tokens, embedding_dim)
        # batch_id : (num_tokens)
        sparse_p = tensor_op.dense2sparse(embedded_p, tokens_p["tokens"]["mask"]) #need to overload indexer for this
        sparse_h = tensor_op.dense2sparse(embedded_h, tokens_h["tokens"]["mask"])
        assert(not torch.any(torch.isnan(sparse_p["data"])))
        assert(not torch.any(torch.isnan(sparse_h["data"])))
        # Shape: (batch_size, classifier_in_dim)
        if return_attention:
            cls_vector, attention_dict = self.encoder(sparse_p,
                                                      sparse_h,
                                                      g_p_embedded,
                                                      g_h_embedded,
                                                      return_attention=return_attention)
        else:
            cls_vector = self.encoder(sparse_p,
                                      sparse_h,
                                      g_p_embedded,
                                      g_h_embedded,
                                      return_attention=return_attention)
        # Shape: (batch_size, num_labels)
        logits = self.classifier(cls_vector)
        assert(not torch.any(torch.isnan(cls_vector)))
        # Shape: (batch_size, num_labels)
        probs = torch.nn.functional.softmax(logits, dim=1)
        # Shape: TensorDict
        output = {'probs': probs}
        if label is not None:
            #print(logits.size(), label.size())
            self.accuracy(logits, label)
            # the two value can be kind of different for numerical isse IMO
            self.entropy(logits, label)
            output['loss'] = torch.nn.functional.cross_entropy(logits, label)
        if return_attention is True:
            output['attentions'] = attention_dict
        return output
    
    # add entropy support for visualizing loss
    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        return {
            "accuracy": self.accuracy.get_metric(reset),
            "entropy": self.entropy.get_metric(reset)["entropy"].item(),
        }
    
    # for interpretation
    def make_output_human_readable(
        self,
        output_dict: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """
        will be used by forward_on_instances() of Model
        which is called by Predictors
        """
        # Take the logits from the forward pass, and compute the label
        # IDs for maximum values
        probs = output_dict['probs'].cpu().data.numpy()
        predicted_id = np.argmax(probs, axis=-1)
        # Convert these IDs back to label strings using vocab
        output_dict['predicted_label'] = [
            self.vocab.get_token_from_index(x, namespace='labels')
            for x in predicted_id
        ]
        # attention is in output_dict already
        # original tokens/glod_label/problemID?
        return output_dict